"""Select2 field implementation module."""

from django.forms import ChoiceField


class ChoiceCallable:
    """Choices wrapper that supports callable choices."""

    def __init__(self, choices):
        """Instanciate with a callable or iterable choices."""
        self.choices = choices

    def __call__(self):
        """Call the callable if necessary and return choices."""
        result = []
        choices = self.choices() if callable(self.choices) else self.choices
        for choice in choices or []:
            if isinstance(choice, (list, tuple)):
                result.append(choice)
            else:
                result.append((choice, choice))
        return result


class Select2ListChoiceField(ChoiceField):
    """Allows a list of values to be used with a ChoiceField.

    Avoids unusual things that can happen if Select2ListView is used for
    a form where the text and value for choices are not the same.
    """

    def __init__(self, choice_list=None, required=True, widget=None,
                 label=None, initial=None, help_text='', *args, **kwargs):
        """Use a list to generate choices in a ChoiceField.

        .. py:param choice_list: The list to use to generate choices or a
        function that returns a list.
        """
        choices = ChoiceCallable(choice_list)

        super(Select2ListChoiceField, self).__init__(
            choices=choices, required=required, widget=widget, label=label,
            initial=initial, help_text=help_text, *args, **kwargs
        )


class Select2ListCreateChoiceField(Select2ListChoiceField):
    """Skips validation of choices so any value can be used."""

    def validate(self, value):
        """Do not validate choices but check for empty."""
        super(ChoiceField, self).validate(value)



"""Select2 view implementation."""

try:
    from collections.abc import Sequence
except ImportError:  # py < 3.10
    from collections import Sequence
from collections import OrderedDict

from dal.views import BaseQuerySetView, ViewMixin

from django import http
from django.core.exceptions import ImproperlyConfigured
from django.db.models import F
from django.utils.translation import gettext as _
from django.views.generic.list import View


class Select2ViewMixin(object):
    """View mixin to render a JSON response for Select2."""

    case_sensitive_create = False

    def get_results(self, context):
        """Return data for the 'results' key of the response."""
        return [
            {
                'id': self.get_result_value(result),
                'text': self.get_result_label(result),
                'selected_text': self.get_selected_result_label(result),
            } for result in context['object_list']
        ]

    def get_create_option(self, context, q):
        """Form the correct create_option to append to results."""
        create_option = []
        display_create_option = False
        if self.create_field and q:
            page_obj = context.get('page_obj', None)
            if page_obj is None or page_obj.number == 1:
                display_create_option = True

            if not self.case_sensitive_create:
                # Don't offer to create a new option if a
                # case-insensitive) identical one already exists
                existing_options = (self.get_result_label(result).lower()
                                    for result in context['object_list'])
                if q.lower() in existing_options:
                    display_create_option = False
            else:
                existing_options = (
                    self.get_result_label(result)
                    for result in context['object_list']
                )
                if q in existing_options:
                    display_create_option = False

        if display_create_option and self.has_add_permission(self.request):
            create_option = [{
                'id': q,
                'text': _('Create "%(new_value)s"') % {'new_value': q},
                'create_id': True,
            }]
        return create_option

    def render_to_response(self, context):
        """Return a JSON response in Select2 format."""
        q = self.request.GET.get('q', None)

        create_option = self.get_create_option(context, q)

        return http.JsonResponse(
            {
                'results': self.get_results(context) + create_option,
                'pagination': {
                    'more': self.has_more(context)
                }
            })


class Select2QuerySetView(Select2ViewMixin, BaseQuerySetView):
    """List options for a Select2 widget."""


class Select2GroupQuerySetView(Select2QuerySetView):
    """List of grouped options for a Select2 widget.

    .. py:attribute:: group_by_related

        Name of the field for the related Model on a One to Many relation

    .. py:attribute:: related_field_name

        Name of the related Model field to run filter against.
    """

    group_by_related = None
    related_field_name = 'name'

    def get_results(self, context):
        """Return the options grouped by a common related model.

        Raises ImproperlyConfigured if self.group_by_name is not configured
        """
        if not self.group_by_related:
            raise ImproperlyConfigured("Missing group_by_related.")

        groups = OrderedDict()

        object_list = context['object_list']
        object_list = object_list.annotate(
            group_name=F(f'{self.group_by_related}__{self.related_field_name}'))

        for result in object_list:
            group_name = getattr(result, 'group_name')
            groups.setdefault(group_name, [])
            groups[group_name].append(result)

        return [{
            'id': None,
            'text': group,
            'children': [{
                'id': self.get_result_value(result),
                'text': self.get_result_label(result),
                'selected_text': self.get_selected_result_label(result),
            } for result in results]
        } for group, results in groups.items()]


class Select2ListView(ViewMixin, View):
    """Autocomplete from a list of items rather than a QuerySet."""

    def get_list(self):
        """Return the list strings from which to autocomplete."""
        return []

    def get(self, request, *args, **kwargs):
        """Return option list json response."""
        results = self.get_list()
        create_option = []
        if self.q:
            results = self.autocomplete_results(results)
            if hasattr(self, 'create'):
                create_option = [{
                    'id': self.q,
                    'text': _('Create "%(new_value)s"') % {
                        'new_value': self.q
                    },
                    'create_id': True
                }]
        return http.JsonResponse({
            'results': self.results(results) + create_option
        }, content_type='application/json')

    def autocomplete_results(self, results):
        """Return list of strings that match the autocomplete query."""
        if all(isinstance(el, list) for el in results) and len(results) > 0:
            return [[x, y] for [x, y] in results if self.q.lower() in y.lower()]
        if all(isinstance(el, tuple) for el in results) and len(results) > 0:
            return [[x, y] for (x, y) in results if self.q.lower() in y.lower()]
        else:
            return [x for x in results if self.q.lower() in x.lower()]

    def results(self, results):
        """Return the result dictionary."""
        if all(isinstance(el, list) for el in results) and len(results) > 0:
            return [dict(id=x, text=y) for [x, y] in results]
        elif all(isinstance(el, tuple) for el in results) and len(results) > 0:
            return [dict(id=x, text=y) for (x, y) in results]
        else:
            return [dict(id=x, text=x) for x in results]

    def post(self, request, *args, **kwargs):
        """Add an option to the autocomplete list.

        If 'text' is not defined in POST or self.create(text) fails, raises
        bad request. Raises ImproperlyConfigured if self.create if not defined.
        """
        if not hasattr(self, 'create'):
            raise ImproperlyConfigured('Missing "create()"')

        text = request.POST.get('text', None)

        if text is None:
            return http.HttpResponseBadRequest()

        text = self.create(text)

        if text is None:
            return http.HttpResponseBadRequest()

        return http.JsonResponse({
            'id': text,
            'text': text,
        })


class Select2GroupListView(Select2ListView):
    """View mixin for grouped options."""

    def get_item_as_group(self, entry):
        """Return the item with its group."""
        group = None
        item = entry

        if isinstance(entry, Sequence) and not isinstance(entry, str):

            entry_length = len(entry)

            if all(isinstance(el, list) for el in entry) and entry_length > 1:
                group, item = entry[0:2]
                return (group, item),
            elif all(isinstance(el, list) for el in entry) and entry_length > 1:
                group, item = entry[0:2]
                return (group, item),

            else:
                if entry_length > 1:
                    group, item = entry[0:2]
                elif entry_length > 0:
                    item = entry[0]

        if not isinstance(item, Sequence) or isinstance(item, str):
            item = (item,)

        return (group, item),

    def get(self, request, *args, **kwargs):
        """Return option list with children(s) json response."""
        results_dict = {}
        results = self.get_list()

        if results:
            if (
                all(isinstance(el, list) for el in results)
                or all(isinstance(el, tuple) for el in results)
            ):
                flat_results = [
                    (group[0], group[1], item[0], item[1]) for entry in results
                    for group, items in self.get_item_as_group(entry)
                    for item in items
                ]

                if self.q:
                    q = self.q.lower()
                    flat_results = [(g, h, x, y) for g, h, x, y in flat_results
                                    if q in y.lower()]
                for group_id, group, item_id, item in flat_results:
                    results_dict.setdefault((group_id, group), [])
                    results_dict[(group_id, group)].append([item_id, item])

                return http.JsonResponse({
                    "results": [
                        {
                            "id": x, "text": y
                        } for x, y in results_dict.pop((None, None), [])
                    ] + [
                        {
                            "id": g[0],
                            "text": g[1],
                            "children": [
                                {"id": x, "text": y} for x, y in l
                            ]
                        }
                        for g, l in results_dict.items()
                    ]
                })

            else:
                flat_results = [(group, item) for entry in results
                                for group, items in self.get_item_as_group(entry)
                                for item in items]

                if self.q:
                    q = self.q.lower()
                    flat_results = [(g, x) for g, x in flat_results
                                    if q in x.lower()]
                for group, item in flat_results:
                    results_dict.setdefault(group, [])
                    results_dict[group].append(item)

                return http.JsonResponse({
                    "results": [
                        {"id": x, "text": x} for x in results_dict.pop(None, [])
                    ] + [
                        {
                            "id": g,
                            "text": g,
                            "children": [
                                {"id": x, "text": x} for x in l
                            ]
                        }
                        for g, l in results_dict.items()
                    ]
                })



"""Helpers for DAL user story based tests."""

import time


class Select2Story(object):
    """Define Select2 CSS selectors."""

    clear_selector = '.select2-selection__clear'
    container_selector = '.select2-container'
    dropdown_selector = '.select2-dropdown'
    input_selector = '.select2-search__field'
    label_selector = '.select2-selection__rendered'
    labels_selector = \
        '.select2-selection__rendered .select2-selection__choice'
    option_selector = '.select2-results__option[aria-selected]'
    widget_selector = '.select2-selection'

    def wait_script(self):
        """Wait for scripts to be loaded and ready to work."""
        tries = 100
        while tries:
            try:
                return self.browser.evaluate_script('yl.registerFunction')
            except:
                time.sleep(.15)
            tries -= 1
        raise Exception('$.select2 was not defined after 15 seconds.')

    def clean_label(self, label):
        """Remove the "remove" character used in select2."""
        return label.replace('\xd7', '')



"""Utils for testing autocompletes."""
from django.apps import apps


class Fixtures(object):
    """Callback for post_migrate to create many objects."""

    def __init__(self, model_name=None):
        """Preset a model name, ie. 'auth.user'."""
        self.model_name = model_name

    def get_model(self, sender):
        """Return either the preset model, either the sender's TestModel."""
        if self.model_name is None:
            return sender.get_model('TModel')
        else:
            return apps.get_model(self.model_name)

    def __call__(self, sender, **kwargs):
        """Call function, calls install_fixtures."""
        model = self.get_model(sender)
        self.install_fixtures(model)

    def install_fixtures(self, model):
        """Install fixtures for model."""
        for n in range(1, 50):
            try:
                model.objects.get(pk=n)
            except model.DoesNotExist:
                model.objects.create(name='test %s' % n, pk=n)


class OwnedFixtures(Fixtures):
    """Fixtures for models with an "owner" relation to User."""

    installed_auth = False

    def install_fixtures(self, model):
        """Install owners and fixtures."""
        User = apps.get_model('auth.user')  # noqa

        self.test, created = User.objects.get_or_create(
            username='test',
            is_staff=True,
            is_superuser=True
        )
        if created:
            self.test.set_password('test')
            self.test.save()

        self.other, created = User.objects.get_or_create(username='other')
        if created:
            self.other.set_password('test')
            self.other.save()

        for n in range(1, 3):
            for u in [self.test, self.other]:
                model.objects.update_or_create(
                    name='test #%s for %s' % (n, u),
                    defaults=dict(owner=u),
                )


fixtures = Fixtures()



"""Default AppConfig for dal_select2."""
from django.apps import AppConfig
# from django.core import checks


class DefaultApp(AppConfig):
    """Default app for dal_select2."""

    name = 'dal_select2'

    def ready(self):
        """Register select2_submodule_check."""



"""Select2 widget implementation module."""

try:
    from functools import lru_cache
except ImportError:
    # py2
    try:
        from backports.functools_lru_cache import lru_cache
    except ImportError:
        lru_cache = None

from dal.widgets import (
    QuerySetSelectMixin,
    Select,
    SelectMultiple,
    WidgetMixin
)

from django import forms
from django.conf import settings
try:
    # SELECT2_TRANSLATIONS is Django 2.x only
    from django.contrib.admin.widgets import SELECT2_TRANSLATIONS
except ImportError:
    SELECT2_TRANSLATIONS = {}
from django.contrib.staticfiles import finders
from django.utils import translation


def is_iterable(x):
    "An implementation independent way of checking for iterables"
    try:
        iter(x)
    except TypeError:
        return False
    else:
        return True


I18N_PATH = 'autocomplete_light/i18n/'


def get_i18n_name(lang_code):
    """Ensure lang_code is supported by Select2."""
    lower_lang = lang_code.lower()
    split_lang = lang_code.split('-')[0]
    # Use the SELECT2_TRANSLATIONS if available
    if SELECT2_TRANSLATIONS:
        if lower_lang in SELECT2_TRANSLATIONS:
            return SELECT2_TRANSLATIONS.get(lower_lang)
        elif split_lang in SELECT2_TRANSLATIONS:
            return SELECT2_TRANSLATIONS.get(split_lang)
    # Otherwise fallback to manually checking if the static file exists
    if finders.find('%s%s.js' % (I18N_PATH, lang_code)):
        return lang_code
    elif finders.find('%s%s.js' % (I18N_PATH, split_lang)):
        return lang_code.split('-')[0]


if lru_cache:
    get_i18n_name = lru_cache()(get_i18n_name)
else:
    import warnings
    warnings.warn(
        'Python2: no cache on get_i18n_name, pip install backports.functools-lru-cache'
    )


class Select2WidgetMixin(object):
    """Mixin for Select2 widgets."""

    def build_attrs(self, *args, **kwargs):
        """Set data-autocomplete-light-language."""
        attrs = super(Select2WidgetMixin, self).build_attrs(*args, **kwargs)
        lang_code = self._get_language_code()
        if lang_code:
            attrs.setdefault('data-autocomplete-light-language', lang_code)
        return attrs

    def _get_language_code(self):
        """Return language code or None."""
        lang_code = translation.get_language()
        if lang_code:
            lang_code = get_i18n_name(
                translation.to_locale(lang_code).replace('_', '-')
            )
        return lang_code

    @property
    def media(self):
        """Return JS/CSS resources for the widget."""
        extra = '' if settings.DEBUG else '.min'
        i18n_name = self._get_language_code()
        i18n_file = (
            '%s%s.js' % (I18N_PATH, i18n_name),
        ) if i18n_name else ()

        return forms.Media(
            js=(
                'admin/js/vendor/select2/select2.full.js',
                'autocomplete_light/autocomplete_light%s.js' % extra,
                'autocomplete_light/select2%s.js' % extra,
            ) + i18n_file,
            css={
                'screen': (
                    'admin/css/vendor/select2/select2%s.css' % extra,
                    'admin/css/autocomplete.css',
                    'autocomplete_light/select2.css',
                ),
            },
        )

    autocomplete_function = 'select2'


class Select2(Select2WidgetMixin, Select):
    """Select2 widget for regular choices."""


class Select2Multiple(Select2WidgetMixin, SelectMultiple):
    """Select2Multiple widget for regular choices."""


class ListSelect2(WidgetMixin, Select2WidgetMixin, forms.Select):
    """Select widget for regular choices and Select2."""


class ModelSelect2(QuerySetSelectMixin,
                   Select2WidgetMixin,
                   forms.Select):
    """Select widget for QuerySet choices and Select2."""


class ModelSelect2Multiple(QuerySetSelectMixin,
                           Select2WidgetMixin,
                           forms.SelectMultiple):
    """SelectMultiple widget for QuerySet choices and Select2."""


class TagSelect2(WidgetMixin,
                 Select2WidgetMixin,
                 forms.SelectMultiple):
    """Select2 in tag mode."""

    def build_attrs(self, *args, **kwargs):
        """Automatically set data-tags=1."""
        attrs = super(TagSelect2, self).build_attrs(*args, **kwargs)
        attrs.setdefault('data-tags', 1)
        return attrs

    def value_from_datadict(self, data, files, name):
        """Return a comma-separated list of options.

        This is needed because Select2 uses a multiple select even in tag mode,
        and the model field expects a comma-separated list of tags.
        """
        values = super(TagSelect2, self).value_from_datadict(data, files, name)
        return ','.join(values)

    def option_value(self, value):
        """Return the HTML option value attribute for a value."""
        return value

    def format_value(self, value):
        """Return the list of HTML option values for a form field value."""
        if not isinstance(value, (tuple, list)):
            value = [value]

        values = set()
        for v in value:
            if not v:
                continue

            v = v.split(',') if isinstance(v, str) else v
            v = [v] if not is_iterable(v) else v
            for t in v:
                values.add(self.option_value(t))
        return values

    def options(self, name, value, attrs=None):
        """Return only select options."""
        # When the data hasn't validated, we get the raw input
        if isinstance(value, str):
            value = value.split(',')

        for v in value:
            if not v:
                continue

            real_values = v.split(',') if hasattr(v, 'split') else v
            real_values = [real_values] if not is_iterable(real_values) else real_values
            for rv in real_values:
                yield self.option_value(rv)

    def optgroups(self, name, value, attrs=None):
        """Return a list of one optgroup and selected values."""
        default = (None, [], 0)
        groups = [default]

        for i, v in enumerate(self.options(name, value, attrs)):
            default[1].append(
                self.create_option(v, v, v, True, i)
            )
        return groups



"""Autocomplete fields for Select2GenericForeignKey choices."""

from dal_queryset_sequence.fields import QuerySetSequenceModelField

from dal_select2_queryset_sequence.views import Select2QuerySetSequenceAutoView
from dal_select2_queryset_sequence.widgets import QuerySetSequenceSelect2

from django.urls import re_path as url

from queryset_sequence import QuerySetSequence


class Select2GenericForeignKeyModelField(QuerySetSequenceModelField):
    """
    Select2GenericForeignKeyModelField class.

    Field that generate automatically the view for the
    QuerySetSequenceSelect2 widget
    """

    def __init__(self, *args, **kwargs):
        """
        Initialize Select2GenericForeignKeyModelField.

        :param args:
        :param model_choice:
            [(Model, 'filter_by', [('forwardfield_name', 'filter_by')]), ],
            List of tuples, for each select2 widget. Model is the model to
            query, 'filter_by' the attribute of the model to apply the filter.
            The list in the tuple is optional, its to forward a field from the
            form to the widget.
        :param field_id: Optional name instead of the automatic one
        :param kwargs:
        """
        model_choice = kwargs.pop('model_choice', None)
        field_id = kwargs.pop('field_id', None)
        self.field_id = field_id if field_id else id(self)
        if model_choice:
            self.model_choice = model_choice
            models_queryset = [model[0].objects.all()
                               for model in model_choice]
            kwargs['queryset'] = QuerySetSequence(*models_queryset)

        super(Select2GenericForeignKeyModelField, self).__init__(*args, **kwargs)

    def as_url(self, form):
        """Return url."""
        url_name = '{}_autocomp_{}'.format(form.__name__, self.field_id)

        forward_fields = {
            forward_tuple[0]
            for field in self.model_choice if len(field) > 2
            for forward_tuple in field[2]
        }
        # a set of the fields to forward.
        # it checks if the 3rd index of the list exists

        self.widget = QuerySetSequenceSelect2(
            url=url_name, forward=forward_fields
        )

        # generate the class to work with multiple gfk
        # (can't work on instance level)
        auto_view = type(
            'Autoview{}{}'.format(form.__name__, self.field_id),
            (Select2QuerySetSequenceAutoView,),
            {'model_choice': self.model_choice}
        )  # send to the view the model and filter list

        return url(
            r'^{}_{}_autocomp$'.format(form.__name__, self.field_id),
            auto_view.as_view(),
            name=url_name
        )



import os

from setuptools import setup, find_packages


# Utility function to read the README file.
# Used for the long_description. It's nice because now 1) we have a top-level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below ...
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(
    name='django-autocomplete-light',
    version='3.12.1-rc1',
    description='Fresh autocompletes for Django',
    author='James Pic',
    author_email='jamespic@gmail.com',
    url='http://django-autocomplete-light.rtfd.org',
    project_urls={
        'Source': 'https://github.com/yourlabs/django-autocomplete-light',
    },
    packages=find_packages('src'),
    package_dir={'': 'src'},
    include_package_data=True,
    zip_safe=False,
    long_description=read('README'),
    license='MIT',
    keywords='django autocomplete',
    install_requires=[
        'django>=3.2',
    ],
    extras_require={
        'nested': ['django-nested-admin>=3.0.21'],
        'tags': ['django-taggit'],
        'genericm2m': ['django-generic-m2m'],
        'gfk': ['django-querysetsequence>=0.11'],
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 3.2',
        'Framework :: Django :: 4.0',
        'Framework :: Django :: 4.1',
        'Framework :: Django :: 4.2',
        'Framework :: Django :: 5.0',
        'Framework :: Django :: 5.1',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ]
)
